############################################
# Purpose: to check whether a code has coherency in the types of subprograms call.
############################################
from AutoRPE.UtilsRPE.Vault import load_vault
from AutoRPE.UtilsRPE.Getter import get_real_in_line
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.Vault as Vault
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error


def get_command_line_arguments():
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-i", "--input-sources", required=True, type=Error.PathType(exists=True, type='dir'),
                        help="Folder of sources produced with ImplementRPEtoSources")
    parser.add_argument("-v", "--vault", default=None,
                        help="Vault obtained with ImplementRPEtoSources")
    parser.add_argument("-w", "--working-precision", default="sp")

    args = parser.parse_args()
    try:
        VariablePrecision.real_id[args.working_precision]
    except KeyError:
        parser.error("Argument -w/--working precision has the wrong format. Use one in ",
                     "[" + ", ".join([*VariablePrecision.real_id]) + "]")
    return args


def code_check_file(file, vault):
    """
    Check that all functions and subroutines in a source code have coherent types.
    """
    module = vault[file.module_name]
    # Case of read_precision module
    if not module:
        return
    module.update_blocks(vault)
    # Loop through lines
    for index, line in enumerate(module.lines):
        # Returns all REAL var on the line and corresponding dummy arguments (check all the intents)
        if line != "" and not (line.count("variable_is_used") or
                               line.count("apply_truncation") or
                               line.count("emulator_variable_precisions")):
            line = BasicFunctions.remove_if_condition(line)
            try:
                dummy_argument, called_argument = get_real_in_line(line, vault, module.blocks[index],
                                                                   intent=["in", "out", "inout"])
            except Error.InterfaceNotFound:
                print(" Interface: ", line)
                continue
            # A subprogram call was found in line
            if dummy_argument is not None:
                for dummy_arg, called_arg in zip(dummy_argument, called_argument):
                    if not VariablePrecision.type_are_equivalent(dummy_arg.type, called_arg.type):
                        print(line)


if __name__ == "__main__":

    args = get_command_line_arguments()

    path_to_input_sources = args.input_sources
    vault_path = args.vault
    # Set working precision
    VariablePrecision.set_working_precision(args.working_precision)
    if vault_path is not None:
        # Loading vault
        vault = load_vault(vault_path)
    else:
        vault = Vault.create_database(path_to_input_sources)

    # Find all files in sources
    all_sources = BasicFunctions.load_sources(path_to_input_sources)
    n_files = len(all_sources)
    # Process Sources
    for index, source_file in enumerate(all_sources):
        print("\rChecking file %i/%i  %20s" % (index + 1, n_files, source_file.module_name), end="")
        code_check_file(source_file, vault)
