import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
import re


def call_to_subroutine(string):
    """ Return the list strings corresponding to the list of calls made in line"""
    ret_val = []
    string = BasicFunctions.remove_if_condition(string)
    # In non preprocessed files calls can be made sequentially, like "call call1; call call2; ..."
    for call in string.split(";"):
        match = re.search(RegexPattern.call_to_subroutine, call, re.I)
        if match:
            ret_val.append(BasicFunctions.close_brackets(string[string.index(match.group().strip()):]))
    return ret_val



def calls_to_functions(string, procedure, vault):
    """ Return the list strings corresponding to the list of functions called in line"""
    if procedure:
        module_name = procedure.module.name
    else:
        module_name = None

    # line without literals to avoid parenthesis inside strings and other ugly stuff
    no_literal_line = BasicFunctions.remove_literals(string)
    ret_val = []

    if re.search(".*function .*", string, re.IGNORECASE):
        return ret_val
    if re.search(".*subroutine.*", string, re.IGNORECASE):
        return ret_val
    # Avoid structures to be identified as subprogram
    try:
        no_literal_line = BasicFunctions.replace_structures(no_literal_line)
    except AttributeError:
        pass
    pattern = RegexPattern.call_to_function
    matches = [m for m in re.finditer(pattern, no_literal_line, flags=re.I)]
    if matches:
        if vault:
            for m in matches:
                fun_name = m.groups()[0]
                # Check is a function stored in vault
                if not vault.get_procedure_by_name(fun_name, module_name):
                    continue
                function_call = BasicFunctions.close_brackets(string[string.index(fun_name):])
                ret_val.append(function_call)
        else:
            for m in matches:
                function_call = BasicFunctions.close_brackets(string[string.index(m.group(1)):])
                ret_val.append(function_call)

    return ret_val


def is_call_to_function(contents, procedure=None, vault=None):
    """
    This function first does a first filter checking that the syntax is similar to a function call and afterwards
    uses one of the two possible approaches, depending on having a vault available or not.
    """
    m = re.match(RegexPattern.call_to_function, contents.strip(), re.I)
    if m:
        if contents == BasicFunctions.close_brackets(contents):
            arguments = find_call_arguments(contents)
            # If argument is a slice then is an array
            for arg in arguments:
                if re.match(r"\w*:\w*", arg):
                    return False
            # In case there are no arguments, for sure it is a function and not an array
            if not arguments:
                return True
            # Otherwise, use one of the two approaches
            if vault is not None:
                return is_call_to_function_vault(contents, procedure, vault)
            else:
                return is_call_to_function_no_vault(contents)
    return False


def find_called_subroutine(string):
    match = re.search(RegexPattern.call_to_subroutine, string, flags=re.I)
    try:
        return match.group(1).strip()
    except AttributeError:
        raise Error.ProcedureNotFound("The string %s seems not to contain a call to sbr. Please check." % string)


def find_called_function(string):
    return string[:string.find("(")].strip()


def strip_call_arguments(_line):
    # Description: Given a string with a comma separated arguments,
    # extract the list of arguments getting rid of the parenthesis
    # Input: String with subroutine declaration
    # Output: List with arguments
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    _line = _line[op[0] + 1:cp[-1]]
    op = [i for i, x in enumerate(_line) if x == "("]
    cp = [i for i, x in enumerate(_line) if x == ")"]
    commas = [i for i, x in enumerate(_line) if x == ","]
    splitted_line = [char for char in _line]
    for n in commas:
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            splitted_line[n] = "#"
    _line = "".join(splitted_line)
    return [_x.strip() for _x in _line.split("#")]


def find_call_arguments(string):
    n_arg, ints = find_argument_intervals(string)
    if n_arg == 0:
        return []
    if ints:
        arguments = [string[interval[0]:interval[1]].strip() for interval in ints]
        return [x for x in arguments if x]
    else:
        return []


def find_argument_intervals(string, _field=None):
    # Given a line with a call to a function, give the index intervals of the different arguments.

    # Mask to avoid scanning inside strings
    mask = BasicFunctions.literal_mask(string)

    # Find all opening parenthesis
    op = [i for i, x in enumerate(string) if x == "(" and not mask[i]]
    # Find all closing parenthesis
    cp = [i for i, x in enumerate(string) if x == ")" and not mask[i]]

    if not op and not cp:
        return 0, None

    # Keep only arguments part
    arguments_text = string[op[0] + 1:cp[-1]]
    # Find remaining parenthesis
    start = op[0] + 1
    end = cp[-1]
    op = [i for i, x in enumerate(arguments_text) if x == "("]
    cp = [i for i, x in enumerate(arguments_text) if x == ")"]
    osp = [i for i, x in enumerate(arguments_text) if x == "[" and not mask[i]]
    csp = [i for i, x in enumerate(arguments_text) if x == "]" and not mask[i]]

    # Find commas

    splitted_line = [c for c in arguments_text]

    commas = []
    mask = BasicFunctions.literal_mask(arguments_text)
    for i, c in enumerate(splitted_line):
        # Had to add this to avoid problems with commas inside strings"
        if c == ",":
            if not mask[i]:
                commas.append(i)

    # For existing commas:
    argument_separating_commas = []
    for n in commas:
        # If comma index n is outside parenthesis (number of openings equal number of closings) substitute
        if len([x for x in op if x < n]) == len([x for x in cp if x < n]):
            if len([x for x in osp if x < n]) == len([x for x in csp if x < n]):
                splitted_line[n] = "#"
                argument_separating_commas.append(n)

    # Displacing the argument_separating_commas to match the original line
    argument_separating_commas = [_x + start + 1 for _x in argument_separating_commas]
    # Adding first and last elements
    argument_separating_commas = [start] + argument_separating_commas + [end + 1]

    # Constructing intervals
    intervals = [(argument_separating_commas[i], argument_separating_commas[i + 1] - 1) for i in
                 range(len(argument_separating_commas) - 1)]
    if _field is not None:
        if len(intervals) <= _field:
            return len(intervals), None
        else:
            return len(intervals), intervals[_field]
    else:
        return len(intervals), intervals


def is_call_to_function_no_vault(contents):
    """
    This function tries to identify if a certain string is a call to a function or it's an array.
    We don't rely on nothing else than the string itself.
    """
    arguments = find_call_arguments(contents)
    for argument in arguments:
        if argument == ":":
            return False
    # TODO: This will wrongly identify arrays with integer incices as functions.
    return True


def is_call_to_function_vault(contents, procedure, vault):
    """
    This function tries to identify if a certain string is a call to a function or it's an array.
    We rely on knowing in which procedure we are and having a vault available.
    """
    pattern = r"[-+]* *(\w+) *\("
    function_name = re.match(pattern, contents).group(1)

    # Its a call to an intrinsic
    if function_name.lower() in BasicFunctions.list_of_intrinsics:
        return True
    if procedure.get_variable_by_name(function_name):
        return False

    # Search for a procedure with the same name
    if vault.get_procedure_by_name(function_name, procedure.module.name):
        return True
    else:
        pass

    # Search in the whole vault a variable with the same name
    try:
        var = vault.get_variable_by_name(function_name, procedure)
        if var.name == var.procedure.name:
            print("Wait a minute!")
        return False
    except Error.VariableNotFound:
        # TODO: Here we are assuming that if it has not been identified neither as a function existing in the vault
        # TODO: nor a variable existing in the vault it corresponds to an external function. In case this assumption is not
        # TODO: True, this should be fixed.
        return True


def get_procedure_inline(line, vault, procedure, only_subroutine=False, only_function=False):

    called_subroutines = []
    called_functions = []

    # Search for calls to subprograms
    if not only_function:
        called_subroutines = get_subroutine(line, procedure, vault=vault)

    if not only_subroutine:
        called_functions = get_function(line, procedure, vault=vault)

    return called_subroutines + called_functions


def get_function(string, procedure, vault):
    from AutoRPE.UtilsRPE.NatureDeterminer import is_function_declaration_statement
    if string.count("::") or is_function_declaration_statement(string):
        return []
    # This will avoid error with regex for functions that has a look ahead with exactly one space
    return get_subprogram(re.sub('\\s\\s+', r' ', string), procedure, vault, RegexPattern.call_to_function,
                          BasicStructures.Function)


def get_subroutine(string, procedure, vault):
    return get_subprogram(string, procedure, vault, RegexPattern.call_to_subroutine, BasicStructures.SubRoutine)


def get_subprogram(string, procedure, vault, regex_pattern, procedure_type):
    import AutoRPE.UtilsRPE.Getter as Getter
    import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
    string = string.strip()
    string = BasicFunctions.remove_literals(string)
    to_return = []
    external_subprogram = []
    is_unknown_external = False
    # Check is not a function or subroutine declaration, to avoid matching with functions syntax
    if re.search(RegexPattern.subprogram_declaration, string, re.MULTILINE | re.IGNORECASE):
        return to_return

    # Avoid structures to be identified as subprogram
    purged_string = BasicFunctions.replace_structures(string)
    matches = [m for m in re.finditer(regex_pattern, purged_string, flags=re.I)]
    if matches:
        for m in matches:
            subprogram_name = m.groups()[0]

            # check is not a variable, a builtin function or an intrinsic
            if not NatureDeterminer.is_subprogram(subprogram_name, procedure, vault):
                continue
            # Get the actual call
            call = re.sub(r'\bcall\b', '', BasicFunctions.close_brackets(string[m.start():]), flags=re.I)
            # Find called arguments
            called_arguments = find_call_arguments(call)

            # We ruled out the possibility of variables or intrinsics: must be a subprogram
            try:
                subprogram = vault.procedures_dictionary[subprogram_name]
            except KeyError:
                # Create an external subprogram to be returned as the error
                subprogram = [procedure_type(subprogram_name, module=None, is_external=True)]
                is_unknown_external = True

            if len(subprogram) > 1:
                subprogram = [s for s in subprogram if s.module.name == procedure.module.name]

            subprogram = subprogram[0]

            if not isinstance(subprogram, procedure_type):
                # Search for the correct subprogram
                if isinstance(subprogram, BasicStructures.Interface):
                    subprogram = Getter.get_interface_from_arguments(called_arguments,
                                                                     interface=subprogram,
                                                                     procedure=procedure,
                                                                     vault=vault)
                    # If the subprogram is not of the correct type, skip the occurrence
                    if not isinstance(subprogram, procedure_type):
                        continue
                elif isinstance(subprogram, BasicStructures.DerivedType):
                    continue
                else:
                    raise Error.ProcedureNotFound("The subprogram %s could not be found", subprogram_name)

            subprogram_call = create_SubprogramCall(call, called_arguments, subprogram, procedure, vault)
            if is_unknown_external:
                external_subprogram.append(subprogram_call)
            else:
                to_return.append(subprogram_call)
    # Raise an error returning all the unknown subprogram calls found in line
    if external_subprogram:
        raise Error.ProcedureNotFound(external_subprogram)
    return to_return


def create_SubprogramCall(call, called_arguments, subprogram, procedure, vault):
    import AutoRPE.UtilsRPE.Getter as Getter
    subprogram_call = BasicStructures.SubprogramCall(call.strip(), subprogram, procedure)

    for index, argument in enumerate(called_arguments):

        key = None
        try:
            if Getter.is_keyword_argument(argument):
                [key, argument] = argument.split("=", 1)
                key = key.strip()
                argument = argument.strip()
                var_type = Getter.get_type_of_contents(argument, procedure=procedure, vault=vault)
            else:
                var_type = Getter.get_type_of_contents(argument, procedure=procedure, vault=vault)
        except Error.VariableNotFound:
            # An external variable has been used
            var_type = "external"

        if not subprogram_call.subprogram.is_external:
            if key is not None:
                dummy_var = subprogram_call.subprogram.get_variable_by_name(key)
            else:
                dummy_var = subprogram_call.subprogram.get_variable_by_position(index)
            subprogram_call.dummy_arguments.append(dummy_var)

        if not var_type == "external":
            try:
                var = Getter.get_variable(argument, procedure=procedure, vault=vault)
            except Error.VariableNotFound:
                var = None
            arg = BasicStructures.Argument(argument, var_type, var)
        else:
            arg = BasicStructures.Argument(argument, type="external", var=None)

        subprogram_call.arguments.append(arg)
    return subprogram_call
