import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.Getter as Getter
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.Inserter as Inserter
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Vault as Vault
import AutoRPE.UtilsRPE.TrackDependency as TrackDependency
import re


# Hardcoded list, this should be solved
list_of_intrinsics_2fix = ["max", "maxloc", "maxval", "min", "minloc", "minval", "epsilon",
                           "isnan", "count", "abs", "mod", "reshape",
                           "atanh", "sqrt", "log",
                           "sum",
                           "erfc",
                           "merge",
                           "dsqrt", "cmplx", "transfer"]


def Implement_RPE(path_to_input_sources, path_to_output_sources, blacklist=[]):

    # Changes all the real declaration for RPE declaration
    implementer = ImplementRPE(path_to_input_sources, path_to_output_sources, blacklist, extension=".f90")

    # Propagate dependency among all variables of type rpe
    TrackDependency.propagate_dependencies(implementer.vault, var_type="rpe_var")

    # Fix the problem the changes implied
    implementer.fix_sources()

    # Insert precision specification into sources
    Inserter.add_sbits_to_rpe_variables(implementer.vault)

    # Add the read_precisions and rpe_emulator use statement into all files (should be accessible for all modules)
    Inserter.insert_use_statement(implementer.vault)

    # Save module txt files to disk
    implementer.save_files()

    # Create the read_precisions.f90 module
    Inserter.create_read_precisions_module(path_to_output_sources, implementer.vault)

    return implementer.vault


class ImplementRPE:
    def __init__(self, path_to_input_sources, path_to_output_sources, blacklist=[], extension=".f90"):
        self.file_list = BasicFunctions.load_sources(path_to_input_sources, extension)
        self.output_path = path_to_output_sources
        self.blacklist = blacklist
        # Replaces real declaration with RPE declaration and save files to output sources
        self.replace_real_with_RPE_declaration()
        # Parse the sources and obtain all info in the output sources
        self.vault = Vault.create_database(path_to_output_sources, extension)

    def save_files(self):
        for module in self.vault.modules:
            module.save_file(self.output_path)

    def replace_real_with_RPE_declaration(self):
        n_of_file = len(self.file_list)
        for index, source_file in enumerate(self.file_list):
            if source_file.filename not in self.blacklist:
                print("\rReplacing real variables %i/%i  %20s" % (index + 1, n_of_file, source_file.filename), end="")

                # Replace real declarations with type(rpe_var)
                # Look at each line for REAL variable declarations and replace it with type(rpe_var)
                for _index, line in enumerate(source_file.lines):
                    _match = re.search(RegexPattern.general_real_declaration, line, flags=re.I)
                    if _match:
                        _match_real_with_prec = re.search(RegexPattern.real_with_precision_declaration, line,
                                                          flags=re.I)
                        if _match_real_with_prec:
                            real_type = _match_real_with_prec.groups()[0]
                            line = line.replace(real_type, "type(rpe_var)")
                            source_file.lines[_index] = line
                        else:
                            _match = re.search(RegexPattern.pure_real_declaration, line, flags=re.I)
                            if _match:
                                real_type = _match.groups()[0]
                            else:
                                raise Exception
                            line = line.replace(real_type, "type(rpe_var)")
                            source_file.lines[_index] = line
                source_file.rebuild_text()

            source_file.save_file(self.output_path)

    def fix_sources(self):
        n_of_modules = len(self.vault.modules)

        for index, module in enumerate(self.vault.modules):
            print("\rFixing modules %i/%i  %20s" % (index + 1, n_of_modules, module.module_name), end="")
            if index == 0:
                # This fix all modules at a time, execute it just once: Remove indistinguishable  interfaces
                self.fix_interfaces()

            # Fix all RPE variables declared with assignation
            self.fix_parameter_declaration(module)

            # Process intrinsic calls
            self.fix_intrinsic_functions(module)

            # Process subroutine and function calls
            self.fix_subprogram_calls(module)

            # Fix calls to type constructor with RPE components
            self.fix_derived_type_constructor(module)

            # Process read and write calls
            self.fix_read_and_write_calls(module)

            # Fix pointer assignations
            self.fix_pointer_assignations(module)

    def fix_parameter_declaration(self, module):
        # If there's an assignation whe have to convert everything to an rpe constructor call
        for _index, line in enumerate(module.lines):
            if re.search(RegexPattern.rpe_parameter_declaration, line, flags=re.I):
                arguments = line.split("::")[1]

                assignation = arguments.split("=")[1].strip()
                content_of_declaration = self._find_kind_of_parameter_declaration(assignation, module.blocks[_index])

                # The argument need to be converted to real
                if not content_of_declaration == "is_a_number":
                    # Take the real part of the rpe expression, to be passed to the RPE constructor
                    assignation_real_part = self.get_assignation_real_part(module.blocks[_index], assignation, line,
                                                                           content_of_declaration)
                    # Substituting the assignation with its real part in line
                    line = line.replace(assignation, assignation_real_part)
                    assignation = assignation_real_part

                # Add the rpe constructor to assignation in line
                module.lines[_index] = Inserter.add_rpe_type_constructor(assignation, line)

        module.rebuild_text()

    def get_assignation_real_part(self, procedure, argument, line, type_of_exception):
        """Return the real part of the rpe expression passed"""
        # The assignation is made with a simple RPE variable
        if type_of_exception == "rpe_var":

            # Take the real part of the RPE var
            return argument + "%val"

        elif type_of_exception == "has_operation":

            # add %val member to rpe variable among the operands
            return self.get_operation_real_part(argument, procedure)
        elif type_of_exception == 'has_intrinsics':

            # Check if the intrinsic argument is an rpe var, and add the %val where needed
            return self.get_intrinsic_real_part(argument, procedure)

        else:
            raise Error.ExceptionNotManaged("This %s is not managed yet" % type_of_exception)

    def fix_interfaces(self):
        all_interfaces = self.vault.interfaces_dictionary
        interfaces_to_fix = [interface for interface in all_interfaces
                             if len(all_interfaces[interface].key) != len(all_interfaces[interface].subprogram)]
        if interfaces_to_fix:
            for interface in interfaces_to_fix:
                interface = all_interfaces[interface]
                for module_procedure in interface.key:
                    if len(interface.key[module_procedure]) > 1:
                        module_procedure_name = interface.key[module_procedure][1]
                        for index, line in enumerate(interface.module.lines):
                            # Matches the name of the interface
                            match = re.search(r'MODULE PROCEDURE .*(\b%s\b)' % module_procedure_name, line)
                            if match:
                                # Remove the offending module procedure, keeps the rest
                                interface.module.lines[index] = "MODULE PROCEDURE" + \
                                                                ",".join([i for i in
                                                                          line.split("MODULE PROCEDURE")[1].split(",")
                                                                          if not i.strip() == module_procedure_name])
                                if interface.module.lines[index].strip() == "MODULE PROCEDURE":
                                    interface.module.lines[index] = ""
                                break
                        if not match:
                            raise Error.ExceptionNotManaged("An indistinguishable interface was not fixed!")

    def fix_derived_type_constructor(self, module):
        for line_index, line in enumerate(module.lines):
            called_derived_type = self._call_to_derived_type_constructor(line, module.module_name)
            if called_derived_type:
                # Find call to type constructor
                constructor_call = BasicFunctions.close_brackets(
                    line[re.search(r'\b%s\b' % called_derived_type.name, line, re.I).start():])
                called_arguments = CallManager.find_call_arguments(constructor_call)
                if not called_arguments:
                    continue

                fixed_constructor_call = constructor_call
                for argument_index, argument in enumerate(called_arguments):
                    argument_type = Getter.get_type_of_contents(argument, module.blocks[line_index], self.vault)
                    dummy_argument = called_derived_type.get_variable_by_position(argument_index)
                    if not dummy_argument:
                        raise AssertionError("Variable not found in position: %i" % argument_index)
                    if VariablePrecision.type_are_equivalent(argument_type, dummy_argument.type):
                        continue
                    elif argument_type in VariablePrecision.real_id and dummy_argument.type == "rpe_var":
                        fixed_constructor_call = Inserter.add_rpe_to_argument(argument, fixed_constructor_call)
                    else:
                        raise AssertionError("Types do not coincide and the case is not considered %s %s" % (
                            argument_type, dummy_argument.type))
                module.lines[line_index] = line.replace(constructor_call, fixed_constructor_call)
        module.rebuild_text()

    def _fix_read_calls(self, module):
        for line_index, line in enumerate(module.lines):
            matches = re.search(RegexPattern.call_to_read, BasicFunctions.remove_if_condition(line), re.I)
            if matches:
                read_call = BasicFunctions.close_brackets(matches.groups()[0])
                # Remove the read call
                read_arguments = BasicFunctions.remove_if_condition(line)[matches.start(1):].replace(read_call, "")
                # Continue just if arguments follow the call
                if not read_arguments:
                    continue
                replacement = read_arguments.replace(" ", "")
                arguments = Splitter.split_read_arguments(replacement)

                # Check the argument type and replace just if is an rpe
                for argument_index, argument in enumerate(arguments):
                    argument_type = Getter.get_type_of_contents(argument, procedure=module.blocks[line_index],
                                                                vault=self.vault)
                    if argument_type == "rpe_var":
                        replacement = Inserter.replace_variable_exact_match(argument, argument + "%val", replacement)

                # Rebuilt the line
                # self.lines[line_index] = read_call + " " + read_arguments
                module.lines[line_index] = line.replace(read_arguments, replacement)

    def _fix_write_calls(self, module):
        pattern = "^[^\!]*\s*write\([^()]*\) (.*)."

        for l_index, line in enumerate(module.lines):
            if re.search("WRITE *\(", line):
                wc = BasicFunctions.close_brackets(line[line.find("WRITE"):])

                line = line.replace(wc, "WRITE()")
                matches = re.search(pattern, line.strip(), re.IGNORECASE)
                if matches:
                    line_container, arguments = Splitter.split_line_container_and_arguments(module.lines[l_index])
                    if not arguments:
                        continue
                    mask = BasicFunctions.literal_mask(line)
                    for argument_index, argument in enumerate(arguments):
                        if mask[line.index(argument) + 1]:
                            continue
                        if re.match(RegexPattern.array_in_write_statement, argument):
                            # For calls like  WRITE(numout,9402) (ilci(ji,jj),ilcj(ji,jj),ji=il1,il2)
                            firs_el = Splitter.split_array_in_write_arguments(argument)[0]
                            argument_type = Getter.get_type_of_contents(firs_el, procedure=module.blocks[l_index],
                                                                        vault=self.vault)
                            if argument_type == "rpe_var":
                                raise Error.ExceptionNotManaged("This part of the code does not yet exist")
                            continue
                        argument_type = Getter.get_type_of_contents(argument, procedure=module.blocks[l_index],
                                                                    vault=self.vault)
                        if argument_type == "rpe_var":
                            arguments[argument_index] = "real( %s )" % argument
                    replacement = ",".join(arguments)
                    module.lines[l_index] = line_container.replace("###TOREPLACE###", replacement)

    def fix_read_and_write_calls(self, module):
        self._fix_read_calls(module)
        self._fix_write_calls(module)

    def fix_pointer_assignations(self, module):

        for line_index, line in enumerate(module.lines):
            if NatureDeterminer.is_pointer_assignment(line):
                # Remove if condition to better search for the pointer target
                clean_line = BasicFunctions.remove_if_condition(line)
                pointer_assignation = re.search(RegexPattern.pointers_assignation, clean_line)
                if pointer_assignation:
                    # Group 1 contains possible pointer attributes
                    pointer = pointer_assignation.group(2).strip()
                    target = pointer_assignation.group(3).strip()
                    pointer_type = Getter.get_type_of_contents(pointer, module.blocks[line_index], self.vault)
                    target_type = Getter.get_type_of_contents(target, module.blocks[line_index], self.vault)
                    if pointer_type in VariablePrecision.real_id and target_type == "rpe_var":
                        module.lines[line_index] = line.replace(target, "%s%%val" % target)

    def fix_intrinsic_functions(self, module):
        # Fix all the intrinsics function that are not overloaded in the RPE library
        for _index, _line in enumerate(module.lines):

            # No need to fix intrinsics in declaration, they have already been fixed
            if re.search(RegexPattern.variable_declaration, _line, re.I):
                continue
            # Cycle through the matches, that can be more than one in a line
            replacements = []
            pieces_to_replace = []
            matches_start_list = []
            mask = BasicFunctions.literal_mask(_line)
            for intrinsic in list_of_intrinsics_2fix:

                # Search for the intrinsics
                fun = r"[ ,(+-=*](%s *)\(" % intrinsic

                matches = re.finditer(fun, _line, flags=re.I)
                matches_start = [m.start(1) for m in matches]

                if matches_start:
                    for start in matches_start:
                        if not mask[start]:
                            matches_start_list.append(start)

            if matches_start_list:
                matches_start_list.sort(reverse=True)
                for start in matches_start_list:
                    # Save the piece that need change
                    pieces_to_replace.append(_line[start:].strip())

                    # Current index
                    i_match = len(pieces_to_replace) - 1
                    # Remove initial minus sign (-)
                    if pieces_to_replace[i_match][0] in ["-", "("]:
                        pieces_to_replace[i_match] = pieces_to_replace[i_match][1:]
                    # Remove brackets if they surrounds the argument
                    pieces_to_replace[i_match] = BasicFunctions.close_brackets(pieces_to_replace[i_match])
                    replacements.append(pieces_to_replace[i_match])

                    _line = _line[:start] + _line[start:].replace(pieces_to_replace[i_match],
                                                                  "ALREADY_FIXED_" + str(i_match))
                    # Split the calls argument
                    arguments = CallManager.find_call_arguments(pieces_to_replace[i_match])
                    for arg in arguments:
                        if arg.count("ALREADY_FIXED_"):
                            if re.match("ALREADY_FIXED_\d+", arg):
                                continue
                            if NatureDeterminer.has_operations(arg):
                                # Split elements, removing the arguments already fixed, adding the others to the loop
                                arguments.extend([el for el in Splitter.split_elements_in_operations(arg) if
                                                  not el.count("ALREADY_FIXED_")])
                                continue
                            if NatureDeterminer.is_logical_comparison(arg):
                                # For example: mask = iszij1 == ALREADY_FIXED_0
                                continue
                            try:
                                # Case in which an intrinsic is used as indexing of an array
                                Getter.get_variable(arg, module.blocks[_index], self.vault)
                            except Error.VariableNotFound:
                                continue
                                # raise Error.ExceptionNotManaged("Check this intrinsic please")

                        argument_type = Getter.get_type_of_contents(arg, module.blocks[_index], self.vault)
                        if argument_type != 'rpe_var':
                            continue

                        if re.match("^\w+$", arg):
                            replacements[i_match] = Inserter.add_val_to_argument(arg, replacements[i_match])
                        else:
                            replacements[i_match] = Inserter.add_real_to_argument(arg, replacements[i_match],
                                                                                  VariablePrecision.real_id["wp"])
            for i_match, (piece_to_replace, replacement) in reversed(
                    list(enumerate(zip(pieces_to_replace, replacements)))):
                # Starting from the end, and substituting just first match we avoid re-substituting matches
                _line = _line.replace("ALREADY_FIXED_" + str(i_match), replacement, 1)
            module.lines[_index] = _line

    def _fix_external_call(self, external_call, line):

        temporal_call = BasicFunctions.remove_if_condition(line)

        called_arguments = external_call.arguments
        # In case the procedure is not found, consider it external
        for argument in called_arguments:
            if argument.type == "rpe_var":
                # If it is not a single variable a cast is safe, it must be intent IN
                if NatureDeterminer.has_operations(argument.name):
                    temporal_call = Inserter.add_real_to_argument(argument.name, temporal_call,
                                                                  VariablePrecision.real_id["dp"])
                else:
                    # Here we add the real member of the RPE type
                    temporal_call = Inserter.add_rpe_real_member_to_argument(argument.name, temporal_call)
        return temporal_call.strip()

    def _fix_arguments(self, line, subprogram_call):
        for s_call in subprogram_call:
            if not s_call.subprogram.is_external:
                for arg, dummy_a in zip(s_call.arguments, s_call.dummy_arguments):
                    # This argument is coherent
                    if VariablePrecision.type_are_equivalent(arg.type, dummy_a.type):
                        continue
                    # Inconsistency between used argument and what the call expects
                    if dummy_a.intent == "in":
                        if dummy_a.type == "rpe_var" and arg.type in VariablePrecision.real_id:
                            line = Inserter.add_rpe_to_argument(arg.name, line)
                        elif dummy_a.type in VariablePrecision.real_id and arg.type == "rpe_var":
                            # For sure there is a dp version of this subprogram, otherwise the code would not work in dp
                            if dummy_a.type == VariablePrecision.real_id['sp']:
                                line = Inserter.add_real_to_argument(arg.name, line,
                                                                     VariablePrecision.real_id['wp'])
                            else:
                                # If dp: if a sp version of the subprogram existed we would have catch it with previous if
                                #        -> mantain dp
                                # If wp: mantain wp
                                line = Inserter.add_real_to_argument(arg.name, line,
                                                                     VariablePrecision.real_id[dummy_a.type])
                    # The value of the argument can be modified by the call, if its an rpe pass the val
                    else:
                        if arg.type == "rpe_var" and dummy_a.type in VariablePrecision.real_id:
                            line = Inserter.add_val_to_argument(arg.name, line)
                        else:
                            raise Error.ExceptionNotManaged("Here we should use temporal variables")
            else:
                line = self._fix_external_call(s_call, line)
        return line

    def fix_subprogram_calls(self, module):

        for _index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.remove_if_condition(line)
            subprogram_call = \
                CallManager.get_procedure_inline(temporal_call, self.vault, module.blocks[_index])
            # If something was found check if can be fixed
            if subprogram_call:
                module.lines[_index] = line.replace(temporal_call, self._fix_arguments(temporal_call, subprogram_call))

    def _find_kind_of_parameter_declaration(self, argument, procedure):
        """Return the nature of the parameter declaration. Depending on the type, the fix will be different"""
        # These are real, but need some manipulation before adding the RPE constructor
        if argument.count(","):
            if argument.count("(/") or argument.count("RESHAPE([") or re.search(r"= *\[.*\]", argument):
                type_of_exception = "is_a_number"
            elif argument.count("rpe_var("):
                type_of_exception = "already_fixed"
        else:
            # Get the type of content
            content_type = Getter.get_type_of_contents(argument, procedure, self.vault)

            # Check for intrinsic first: they return a real, but their argument need to be fixed if its an RPE
            if NatureDeterminer.has_intrinsics(argument):
                type_of_exception = 'has_intrinsics'
            # If it doesn't have intrinsic and is real or integer, then is a number: just add and RPE constructor
            elif content_type in list(VariablePrecision.real_id.keys()) + ['integer']:
                return "is_a_number"
            # Need to split operand, find rpe_variable and take their real part
            elif NatureDeterminer.has_operations(argument):
                type_of_exception = 'has_operation'
            # Just a %val will be added
            else:
                type_of_exception = "rpe_var"

        if not type_of_exception:
            raise Error.ExceptionNotManaged("Unknown type of parameter declaration")
        return type_of_exception

    def get_operation_real_part(self, argument, procedure):
        # Get the elements in operation
        split_elements = Splitter.split_elements_in_operations(argument)
        argument_real_part = argument
        for element in split_elements:
            # If is an RPE get the real part
            element = element.strip()

            if Getter.get_type_of_contents(element, procedure, self.vault) == 'rpe_var':
                element_real_part = " %s%%val " % element
            # In case of intrinsic we need to fix the argument, because it has to be real
            elif NatureDeterminer.has_intrinsics(element):
                element_real_part = self.get_intrinsic_real_part(element, procedure)
            else:
                continue
            argument_real_part = Inserter.replace_variable_exact_match(element, element_real_part,
                                                                       argument_real_part)
        return argument_real_part

    def get_intrinsic_real_part(self, element, procedure):
        # We need to fix ALL intrinsic, because their overload is not available at declaration time
        intrinsic_argument = NatureDeterminer.strip_from_intrinsic(element)
        rpe_element = []
        for el in intrinsic_argument:
            if Getter.get_type_of_contents(el, procedure, self.vault) == 'rpe_var':
                # Get the real part of the RPE variable
                if NatureDeterminer.has_operations(el):
                    # List rpe element involved in the operation
                    rpe_element.extend([op for op in Splitter.split_elements_in_operations(el)
                                        if Getter.get_type_of_contents(op, procedure, self.vault) == "rpe_var"])

                else:
                    rpe_element.append(el)
        for el in rpe_element:
            element = Inserter.replace_variable_exact_match(el, el + "%val", element)
        return element

    def _call_to_derived_type_constructor(self, line, module_name):
        line = BasicFunctions.remove_literals(line)
        matches = re.finditer(RegexPattern.assignation_to_DT, line, flags=re.I)
        if matches:
            for m in matches:
                fun_name = m.groups()[0]
                try:
                    procedure = self.vault.get_procedure_by_name(fun_name, module_name)
                    if isinstance(procedure, BasicStructures.DerivedType):
                        return procedure
                except Error.ProcedureNotFound:
                    pass
        return None
