import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import warnings
import re
import itertools
from os.path import join, isfile


def insert_use_statement(vault):
    # Adding module rp_emulator (in theory adding this to par_kind.f90 should be enough,
    # but theory fails and compiler complains so I'll add it everywhere
    for module in vault.modules:
        module.lines.insert(1, "\nUSE rp_emulator\nUSE read_precisions\n")
        module.rebuild_text()


def create_read_precisions_module(path_to_sources, vault=None, target_module=None):
    from os.path import dirname, abspath
    print("\n\nCreating module to read precision namelist...")

    # Create module
    rpe_vars = [var for var in vault.variables if var.type == "rpe_var"]
    assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
    ids = [var.id for var in assigned_rpe_vars]
    # Path to template is hardcoded
    package_directory = dirname(abspath(__file__))
    filename = package_directory + "/../AdditionalFiles/read_precisions_template"

    with open(filename, "r") as input_file:
        template = input_file.read()
    module = template.replace("%NUMBER_OF_RPE_VARS%", "%i" % max(ids))
    with open(join(path_to_sources, "read_precisions.f90"), "w") as output_file:
        output_file.write(module)

    # Put call into target module
    if target_module:
        if isfile(join(path_to_sources, target_module)):
            with open(join(path_to_sources, target_module), "r") as input_file:
                original_source = input_file.read()

            lines = original_source.split("\n")
            lines.insert(1, "use read_precisions")
            for index, line in enumerate(lines):
                if line.lower().strip() == "private":
                    lines.pop(index)
            original_source = "\n".join(lines)

            with open(join(path_to_sources, target_module), "w") as output_file:
                output_file.write(original_source)
        else:
            warnings.warn("File %s not found in sources.\nUnable to insert read_precisions module")


def add_sbits_to_rpe_variables(vault):
    print("\n\nAssigning significant bits...")

    def condition(v):
        return v.type == "rpe_var" and not v.is_parameter and not v.name.count("tmp_var_")

    rpe_variables = [var for var in vault.variables if condition(var) and var.dimension is not None]
    allocatable, non_allocatable = Splitter.split_vars_by_allocatable(rpe_variables)
    var_counter = Counter()

    # Allocatable
    for variable in allocatable:
        insert_variable_precision_specification_to_allocatable(variable, var_counter)

    # Non allocatable that appear in a namelist
    na_in_namelist, na_not_in_namelist = Splitter.split_by_appearance_in_namelist(non_allocatable)
    for variable in na_in_namelist:
        insert_variable_precision_specification_to_namelist_parameter(variable, var_counter, vault)
    in_main, not_in_main = Splitter.split_vars_by_main_or_not(na_not_in_namelist)

    # Variables inside routines
    for variable in not_in_main:
        if not isinstance(variable.procedure, BasicStructures.DerivedType) and variable.intent != "in":
            insert_variable_precision_to_routine_variable(variable, var_counter)

    # Adding sbits assignation and apply_truncation after routine calls
    insert_variable_precision_after_subroutine_call(vault, var_counter)


def _add_argument_truncation(line, subprogram_call, counter):
    # Check the intent and type of the arguments
    if_condition = line[0]
    call_to_subroutine = line[1]
    for s_call in subprogram_call:
        variables_to_truncate = []
        if not s_call.subprogram.is_external:
            for arg, dummy_a in zip(s_call.arguments, s_call.dummy_arguments):
                if dummy_a.intent in ["out", "inout"]:
                    # The var can not be None: since the intent is out, must be a single var
                    if arg.variable is not None:
                        if arg.variable.type == "rpe_var":
                            # If var was not yet assigned an id, assign one and update total number of mutable var
                            if arg.variable.id is None:
                                arg.variable.id = counter.count
                                arg.variable.mutable = True
                                counter.up()

                            variables_to_truncate.append([arg.name.replace("%val", "").strip(), arg.variable.id])
                    else:
                        raise Error.IntentOutError("The dummy argument %s of procedure %s has intent %s but is used "
                                                   "like %s "
                                                   % (dummy_a.name, dummy_a.procedure.name, dummy_a.intent, arg.name))
        else:
            # If the routine is external
            for arg in s_call.arguments:
                # If the argument is not a single var, it can not be intent out/inout, so %sbits will stay constant
                if arg.name.count("%val") and not NatureDeterminer.has_operations(arg.name):
                    # If var was not yet assigned an id, assign one and update total number of mutable var
                    if arg.variable.id is None:
                        arg.variable.id = counter.count
                        arg.variable.mutable = True
                        counter.up()
                    # Add it to the array of variable to truncate
                    variables_to_truncate.append([arg.name.replace("%val", "").strip(), arg.variable.id])

        for var, id_num in variables_to_truncate:
            call_to_subroutine += "\n"
            call_to_subroutine += if_condition + "    %s%%sbits = emulator_variable_precisions(%i) \n" % (var, id_num)
            call_to_subroutine += if_condition + "    CALL apply_truncation(%s)" % var
    return call_to_subroutine


def insert_variable_precision_specification(variable, module, insertion_line, condition="", member_name=None):
    # Get name of variable, or of member if specified
    var_name = member_name if member_name else variable.name

    # Add apply_truncation
    new_line = condition + "    CALL apply_truncation(%s)" % var_name
    module.lines.insert(insertion_line, new_line)

    # Add sbits    
    new_line = condition + "    %s%%sbits = emulator_variable_precisions(%i)" % (var_name, variable.id)
    module.lines.insert(insertion_line, new_line)

    # Add variable usage specification
    new_line = condition + "    variable_is_used(%i) = .true." % variable.id
    module.lines.insert(insertion_line, new_line)

    module.rebuild_text()


def insert_variable_precision_specification_to_allocatable(variable, counter_object):
    for module in [variable.module] + variable.procedure.module.used_by:
        for line_index, line in enumerate(module.lines):
            if re.search(RegexPattern.allocation_variable_name % variable.name, line, re.I):
                if variable.is_member_of:
                    # Get the arguments from ALLOCATE, avoiding any possible extra call or IF statement inline
                    arguments = CallManager.find_call_arguments(BasicFunctions.remove_if_condition(line))
                    for arg in arguments:
                        if arg.count(variable.name):
                            # Remove array arguments (if any) and get only the member and var: foo%var(2:3) --> foo%var
                            member_name = arg.split(variable.name)[0] + variable.name
                            break
                else:
                    # Check that we did not match a member with the same name: a_i will also match %a_i
                    if re.search(RegexPattern.allocation_member_name % variable.name, line.replace(" ",""), re.I):
                        continue
                    member_name = None
                variable.id = counter_object.count
                variable.mutable = True
                counter_object.up()
                insert_variable_precision_specification(variable, module, line_index + 1, member_name=member_name)
                return


def insert_variable_precision_specification_to_namelist_parameter(variable, counter_object, vault):
    module = vault[variable.appears_in_namelist[0]]
    read_index = -1
    for index, line in enumerate(module.lines):
        pattern3 = r"read *\(.*\b%s\b" % variable.appears_in_namelist[1]
        if re.search(pattern3, line, re.I):
            read_index = index
            break
    variable.id = counter_object.count
    variable.mutable = True
    counter_object.up()
    insert_variable_precision_specification(variable, module, index + 1)


def insert_variable_precision_to_routine_variable(variable, counter_object):
    module, first_use_line = find_first_use(variable)
    if module:
        variable.id = counter_object.count
        variable.mutable = True
        counter_object.up()
        _line = module.lines[first_use_line]
        if re.search(r'^\s*if\s*\(', _line, re.I):
            condition = BasicFunctions.close_brackets(_line)
        else:
            condition = ""
        # In case the variable is member of a structure, remove last indexing from expression and pass this as var name
        member_name = None
        if variable.is_member_of:
            member_name = BasicFunctions.remove_if_condition(_line).split(variable.name)[0] + variable.name
        insert_variable_precision_specification(variable, module, first_use_line, condition, member_name)
    else:
        warnings.warn("Variable %s not found %s in routine %s" % (
            variable.name, variable.procedure.module.name, variable.procedure.name))


def insert_variable_precision_after_subroutine_call(vault, counter):
    for module in vault.modules:
        module.update_blocks(vault)
        for _index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.split_if_condition(line)
            try:
                subroutine_calls = \
                    CallManager.get_procedure_inline(temporal_call[1], vault, module.blocks[_index],
                                                     only_subroutine=True)
            except Error.ProcedureNotFound as procedure:
                if procedure.args[0][0].subprogram.name == "apply_truncation":
                    continue
                raise Error.ProcedureNotFound(procedure.args)
            # If something was found check is using a rpe variable
            if subroutine_calls:
                module.lines[_index] = line.replace(temporal_call[1],
                                                    _add_argument_truncation(temporal_call, subroutine_calls, counter))
        module.rebuild_text()


# End fo rpe insertion


def add_val_to_argument(argument, string):
    string = replace_variable_exact_match(argument, argument + "%val", string)
    return string


def add_rpe_real_member_to_argument(argument, string):
    if NatureDeterminer.is_an_hardcoded_array(argument):
        for arg in Splitter.split_hardcoded_array(argument):
            string = replace_variable_exact_match(arg, arg + "%val", string)
        return string
    return replace_variable_exact_match(argument, argument + "%val", string)


def add_rpe_to_argument(argument, string):
    # Check we are not casting more than once
    if not string.count("rpe(" + argument + ")") == string.count(argument):
        if NatureDeterminer.is_an_hardcoded_array(argument):
            string = add_rpe_to_array_arguments(argument)
        else:
            string = replace_variable_exact_match(argument, "rpe(" + argument + ")", string)
    return string


# Transforms a number in an rpe variables =>  1.0 in rpe( 52, 1.0). Used to fix parameters assignation
def add_rpe_type_constructor(argument, string):
    # Deal with cases of array or reshape functions
    if argument.count("RESHAPE") or re.search(r"= *\[.*\]", argument):
        # List the arguments inside the RESHAPE function and avoid adding the cast to dimensions
        argument = CallManager.find_call_arguments(argument)[0]
        return replace_variable_exact_match(argument, add_rpe_to_array_arguments(argument), string)
    if argument.count("(/"):
        # Add rpe constructor to all numbers inside the array
        return replace_variable_exact_match(argument, add_rpe_to_array_arguments(argument), string)

    return replace_variable_exact_match(argument, "rpe_var( 52, " + argument + ")", string)


def add_real_to_argument(argument, string, precision=VariablePrecision.real_id["wp"]):
    # Check we are not casting more than once or fixing an arg already fixed
    precision = ", " + precision
    if string.count("real(" + argument + precision + ")") == string.count(argument):
        return string

    # # Avoid add real to all array element if is an array with a loop
    # if NatureDeterminer.is_an_hardcoded_array(argument) and \
    #         not NatureDeterminer.is_an_hardcoded_array_with_loop(argument):
    #     new_arg = add_real_to_array_arguments(argument)
    #
    # else:
    new_arg = "real(" + argument + precision + ")"

    string = replace_variable_exact_match(argument, new_arg, string)
    return string


def add_real_to_array_arguments(argument):
    simple_array = argument.replace("(/", "")
    simple_array = simple_array.replace("/)", "")
    clean_numbers = CallManager.find_call_arguments("(" + simple_array + ")")
    # Open the array
    ret_val = "(/"
    for number in clean_numbers:
        # Cast all array numbers
        ret_val += "real(" + number + " , wp), "
    # Remove last comma and closes array
    ret_val = ret_val[:-2] + "/)"
    return ret_val


def add_rpe_to_array_arguments(argument):
    simple_array = argument.replace("(/", "")
    simple_array = simple_array.replace("/)", "")
    clean_numbers = CallManager.find_call_arguments("(" + simple_array + ")")
    # Open the array
    ret_val = "(/"
    for number in clean_numbers:
        # Cast all array numbers
        ret_val += "rpe_var( 52, " + number + " ), "
    # Remove last comma and closes array
    ret_val = ret_val[:-2] + "/)"
    return ret_val


def replace_variable_exact_match(piece_to_replace, replacement, string):
    # Prepare the string for a regex, i.e escaping brackets or wildcards
    piece_to_replace = RegexPattern.escape_string(piece_to_replace)

    if not piece_to_replace.strip()[-1] == ".":
        piece_to_replace = piece_to_replace + r"(?!%val)"
        piece_to_replace = piece_to_replace + r"(?! *\()"
    # Before performing the substitution remove strings to avoid substitution inside them
    mask = BasicFunctions.literal_mask(string)
    literal_part = []
    n_of_literal = -1

    # Transforms the string in a list of char (strings are immutable and cannot be modified)
    string = list(itertools.chain.from_iterable([list(piece) for piece in list(string)]))
    for index, i in enumerate(mask):
        if i:
            # A new literal is beginning, add a list to store its value
            if index == 0 or not mask[index - 1]:
                n_of_literal += 1
                literal_part.append([])
            # The char is stored for late substitution
            literal_part[n_of_literal].append(string[index])
            # The char is overwrite with a placeholder
            string[index] = "#"

    # Reconvert the list to string and substitute the fix
    string = "".join(string)
    # Check the argument is passed with the same name as the keyword
    keyword = re.search(r'(%s\s*=\s*)%s' % (piece_to_replace, piece_to_replace), string, re.I)
    if keyword:
        string = list(itertools.chain.from_iterable([list(piece) for piece in list(string)]))
        # avoid casting the keyword
        len_keyword = len(keyword.group(1))
        n_of_literal += 1
        literal_part.append([])
        for i in range(keyword.start(), keyword.start() + len_keyword):
            literal_part[n_of_literal].append(string[i])
            string[i] = "#"

        string = "".join(string)

    string = re.sub(piece_to_replace, replacement, string)

    # Substitutes one at a time the placeholder for the char
    for index, substitution in enumerate(literal_part):
        string = re.sub("#+", "".join(literal_part[index]), string, 1)

    if not string.count(replacement):
        warnings.warn(" The requested substitution %s was not possible in string %s" % (replacement, string))

    return string


def _remove_keywords(arguments):
    """
        In case there's a keyword argument, return only the value.
        """
    patt = "[a-zA-Z_0-9]* *=(.*)"
    for index, argument in enumerate(arguments):
        m = re.match(patt, argument.strip())
        if m:
            arguments[index] = m.group(1).strip()
    return arguments


def find_call_in_original_file(subprogram, argument, argument_index, _file, target_type="dp", fix=False):
    """
        Open a file and look for a specific line that contains the subprogram and the arguments.
    """
    # TODO: Probably shouldn't use sp and wp since those are model specific.
    if target_type == "sp":
        target_type = "wp"
    # Add a cast to the proper type
    new_argument = "REAL(%s, %s)" % (argument, target_type)
    # Fix in case the argument already had a cast, in which case we'll only change the type of the cast
    if new_argument.count("REAL(REAL("):
        patt = "REAL\((.*),(.*)\)"
        m = re.search(patt, argument)
        new_argument = "REAL(%s, %s)" % (m.group(1), target_type)
    # Fix the pattern to account for regex rules
    pattern = r"\b%s\b####$$$$%s" % (subprogram, argument)
    pattern = pattern.replace("(", "\(")
    pattern = pattern.replace(")", "\)")
    pattern = pattern.replace("+", "\+")
    pattern = pattern.replace("*", "\*")
    pattern = pattern.replace("####", ".*")
    pattern = pattern.replace("$$$$", r"(\b|, *)")
    pattern += r"(\b|[, )])"

    # Initialize found to False
    found = False

    # Open file
    with open(_file, "r") as f:
        lines = [l for l in f]
        for line_index, line in enumerate(lines):
            if line.lower().count("subroutine") or line.lower().count("function"):
                continue

            # In case there's a call to the target routine with a line continuation, fix it and call again the function
            if line.lower().count(subprogram):
                # The line contains the subprogram name but the pattern doesn't match
                line = BasicFunctions.remove_comments(lines[line_index]).strip()
                # Check if it is a call with split args
                if line and line.strip()[-1] == "&":
                    line = BasicFunctions.merge_line_range(lines, line_index)
                    m = re.search(pattern, line, re.I)
                    if m:
                        # It merge the split call in one single line
                        BasicFunctions.merge_line_in_file(_file, line_index)
                        # Now the function can properly fix the call
                        return find_call_in_original_file(subprogram, argument, argument_index, _file, target_type, fix)
            else:
                continue

            # Look for the call to the routine with the proper arguments
            m = re.search(pattern, line, re.I)
            if m:
                calls = CallManager.call_to_subroutine(line)

                if len(calls) > 1:
                    print("Now what?")
                arguments = _remove_keywords(CallManager.find_call_arguments(calls[0]))
                # Check if the argument is placed in the proper place
                try:
                    if arguments.index(argument) != argument_index:
                        continue
                except ValueError:
                    # It seems that this can happen if the exact same combination
                    # of routine argument has been fixed before.
                    continue
                found = True
                # TODO: Need to account for the possibility of having the same argument in different positions.
                new_line = replace_variable_exact_match(argument, new_argument, line)
                lines[line_index] = new_line + "\n"

    # In case we found the call overwrite the file with the fix
    # Otherwise return False
    if found:
        if fix:
            code = "".join(lines).strip()
            with open(_file, "w") as output:
                output.write(code)
        return True
    else:
        return False




## PRIVATE

class Counter:
    def __init__(self):
        self.count = 0

    def up(self, how_much=1):
        self.count += how_much

def find_first_use(variable):
    procedure = variable.procedure
    module = procedure.module

    # pattern = '(^|\W+)%s(\W+|$)' % variable.name
    pattern = r'\W*%s\W*' % variable.name
    searching = False
    loop_level = 0
    loop_start = None
    in_where_statement = False
    where_statement_start = None

    # Scroll all the lines
    for line_index, line in enumerate(module.lines):
        line = line.strip().lower()
        # Until it finds the subroutine to which the variables belongs (routine.name)
        if not searching:
            match = re.search(r"(subroutine|function)\s+\b%s\b" % procedure.name, line, re.IGNORECASE)
            if match:
                searching = True
        # In the subroutine
        else:
            # Scroll all the lines searching for the first use
            if re.search(pattern, line, re.IGNORECASE):
                # skip declarations: bad workaround, but avoids false positive that will block compilation.
                if line.count("::") or line.lower().count("implicit none"):
                    continue
                # Skip the first line
                elif line.count("subroutine"):
                    # Reach the end: the variable has not been used
                    if line.count("end subroutine"):
                        break
                    continue
                # Skip import statements
                elif line.find("use ") == 0:
                    continue
                # Other statement
                elif line.find("data ") == 0:
                    continue
                else:
                    # Check the variable is not mentioned in a ".NOT.PRESENT" statement
                    if _check_if_condition(line, variable.name):
                        continue
                    # The variable is found
                    if in_where_statement:
                        # Inside a where statement
                        return module, where_statement_start
                    else:
                        # In a (nested) loop
                        if loop_level > 0:
                            return module, loop_start
                        # On a normal line
                        else:
                            return module, line_index
            elif line.count("end subroutine"):
                break
            # Whenever a 'do loop' starts increment nesting index
            elif re.search(r"^ *do *[\w]* *=", line, re.IGNORECASE):
                if loop_level == 0:
                    loop_start = line_index
                loop_level += 1
            # Whenever a 'do loop' ends decrement nesting index
            # This way match bot the expression 'end do' and 'enddo' but excludes matches like 'enddo loop_name'
            elif re.search(r"^ *end\s*do", line, re.IGNORECASE):
                loop_level -= 1
            # Do the same for where statement
            elif line.find("where") == 0:
                in_where_statement = True
                where_statement_start = line_index
            elif line.find("end where") == 0:
                in_where_statement = False
    return None, None


def _check_if_condition(line, var_name):
    condition, line = BasicFunctions.split_if_condition(line)
    if condition:
        arg = CallManager.find_call_arguments(condition)
        for i_arg in arg:
            if re.search(r"\.NOT\.\s*PRESENT\s*\(\s*%s" % var_name, i_arg, re.I):
                return True
    return False


def _in_which_type_definition(_line, type_def):
    _line = _line.strip().lower()
    _line = BasicFunctions.remove_literals(_line)
    _pattern_type = "^ *type[^(].*::(.*)"
    _pattern_type_alt = "^ *type[^(](.*)"
    _pattern_type_end = "^ *end *type"

    # We are changing block
    if type_def is None:
        m = re.match(_pattern_type, _line, re.I)
        if m:
            type_name = m.group(1).strip()
            return type_name
        n = re.match(_pattern_type_alt, _line, re.I)
        if n:
            type_name = n.group(1).strip()
            return type_name
    else:
        m = re.search(_pattern_type_end, _line, re.I)
        if m:
            return None
    return type_def
