import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.ObtainSourceFileInfo as ObtainSourceFileInfo
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Preprocessor as Preprocessor
import AutoRPE.UtilsRPE.Error as Error
import warnings


def load_vault(path):
    import pickle
    try:
        return pickle.load(open(path, "rb"))
    except IOError:
        return None


def create_database(path_to_sources, extension=".f90"):
    # Initialize an empty database
    vault = Vault()
    # Load all the files in the provided path with extension
    file_list = BasicFunctions.load_sources(path_to_sources, extension)
    # Removes comment and join split lines
    Preprocessor.clean_sources(file_list)
    # Fill database with info from input files
    ObtainSourceFileInfo.parse_sources(file_list, vault, extension)
    return vault


class Vault:
    def __init__(self):
        self.modules = []
        self.variables = []
        self.variables_dictionary = {}
        self.procedures_dictionary = {}
        self.interfaces_dictionary = {}
        self.derived_types = []

        self.indexed_variables = None

    def __getitem__(self, item):
        for module in self.modules:
            if module.name == item:
                return module

    def add_procedure(self, procedure):
        try:
            self.procedures_dictionary[procedure.name] += [procedure]
        except KeyError:
            self.procedures_dictionary[procedure.name] = [procedure]

        if not isinstance(procedure, BasicStructures.Interface):
            for variable in procedure.variables:
                try:
                    self.variables_dictionary[variable.name] += [variable]
                except KeyError:
                    self.variables_dictionary[variable.name] = [variable]

    def index_variables(self):
        self.indexed_variables = {}
        for variable in self.variables:
            self.indexed_variables[variable.id] = variable

    def get_variable_from_namelist_by_name(self, var_name, namelist):
        var = [var for var in self.variables if var.name == var_name and var.appears_in_namelist and namelist in var.appears_in_namelist]
        if len(var) > 0:
            return var[0]
        else:
            raise Error.VariableNotFound

    def get_variable_by_name(self, var_name, procedure):
        # Search just among the accessible variables
        try:
            private_variables = procedure.accessible_var[var_name]

            if len(private_variables) == 1:
                return private_variables[0]
            # A variable has been declared both as global (procedure.name == main ) and as private procedure variable
            elif len(private_variables) > 1:
                # We check hierarchically
                try:
                    shadow_var = [v for v in private_variables if v in procedure.variables]
                    warnings.warn("A global variable %s has been shadowed " % private_variables[0].name)
                    return shadow_var[0]
                except IndexError:
                    try:
                        shadow_var = [v for v in private_variables if v in procedure.module.variables]
                        warnings.warn("A global variable %s has been shadowed " % private_variables[0].name)
                        return shadow_var[0]
                    except IndexError:
                        try:
                            shadow_var = [v for v in private_variables if v in procedure.module.accessible_var]
                            warnings.warn("A global variable %s has been shadowed " % private_variables[0].name)
                            return shadow_var[0]
                        except IndexError:
                            raise Error.VariableNotFound
        except KeyError:
            try:
                # Search among all the variables stored in vault
                global_variables = self.variables_dictionary[var_name]
            except KeyError:
                raise Error.VariableNotFound
            # Just one var exists with this name
            if len(global_variables) == 1:
                return global_variables[0]
            elif len(global_variables) > 1:
                # A variable is used from a module indirectly imported

                var = [c for c in global_variables if c in procedure.module.indirect_accessible_var]
                return var[0]

    def get_variable_by_id(self, var_id):
        # Purpose: Retrieve a variable by id

        # Few workarounds in case the vault does not have the indexed_variables
        try:
            self.indexed_variables
        except AttributeError:
            self.indexed_variables = None
        if not self.indexed_variables:
            self.index_variables()

        return self.indexed_variables[var_id]

    def get_procedure_by_name(self, item, module_name=None):
        # Purpose: Retrieve a procedure by name
        # item = item.lower()
        try:
            procedure = self.procedures_dictionary[item]
            if len(procedure) > 1 and module_name is not None:
                # this is safe until they start to define functions inside functions
                return [proc for proc in procedure if proc.module.name == module_name][0]

            return self.procedures_dictionary[item][0]
        except KeyError:
            return None

    def get_derived_type(self, item):
        # Purpose: Retrieve a derived type by name
        for dt in self.derived_types:
            if dt.name.lower() == item.lower():
                return dt

    def expand_use_statement(self):
        for module in self.modules:
            # Store use statement of my used modules
            additional_used_modules = []
            for um in module.uses:
                for var in um.variables:
                    module.accessible_var.append(var)
                if um.uses:
                    # Add module used by my use statement
                    for umu in um.uses:
                        # Exclude modules already present
                        if umu not in additional_used_modules + module.uses:
                            additional_used_modules.append(umu)

            # On the first loop, used_modules will be: the use statement made by my used modules
            # On the second loop used_modules will be: the use statement made by their used modules
            # Until I recursively visit all the used modules
            while not set(additional_used_modules) == set(module.uses):
                used_modules = []
                for aum in additional_used_modules:
                    module.uses.append(aum)
                    if aum.uses:
                        used_modules.extend(um.uses)
                additional_used_modules = set(used_modules + module.uses)

        # Add public var of used modules to available var
        for module in self.modules:
            for used_module in module.uses:
                module.indirect_accessible_var.extend(list(set(used_module.variables) - set(module.accessible_var)))

        # Loop through the list to see which modules use mine
        for module in self.modules:
            for mu in module.uses:
                if module not in mu.used_by:
                    mu.used_by.append(module)

        # Add info when a Func or a Sbr is using a use statement:
        subprograms = [p for p in self.procedures_dictionary.values() if
                       isinstance(p[0], BasicStructures.SubRoutine) or
                       isinstance(p[0], BasicStructures.Function)]

        for subprogram in subprograms:
            # Subprograms with the same name are stored in a list associated with the same key
            for sub in subprogram:
                list_of_accessible_variables = []
                if len(sub.use_only_variables):
                    for index, module in enumerate(sub.use_only_variables):
                        if module.count("only"):
                            used_var = module.split("only")[1].replace(":", "").strip()
                            module = module.split("only")[0].replace(",", "").strip()
                            for var in used_var.split(","):
                                used_module = [m for m in self.modules if m.name == module][0]
                                variable = [v for v in used_module.variables if v.name == var.strip()]
                                if variable:
                                    list_of_accessible_variables.append(variable[0])
                        else:
                            used_module = [m for m in self.modules if m.name == module]
                            if used_module:
                                list_of_accessible_variables.extend(used_module[0].variables)
                    if list_of_accessible_variables:
                        sub.use_only_variables = list_of_accessible_variables
                    else:
                        sub.use_only_variables = []
                sub.fill_dictionary_of_variable()
                sub.fill_dictionary_of_accessible_var()

    def expand_derived_types(self):
        for derived_type in self.derived_types:
            if derived_type.extended_structure is not None:
                # Save mother structure in extension
                struct_to_extend = [s for s in self.derived_types if s.name == derived_type.extended_structure][0]
                struct_to_extend.extension = derived_type
                # Save extension the mother structure
                derived_type.extension = struct_to_extend
            # Now that everything is stored, we can fill the accessible variables
            derived_type.fill_dictionary_of_accessible_var()

    def store_external_data_structures(self):
        # Derived data types to avoid
        avoid_types = [dt.name.lower() for dt in self.derived_types] + ['char', 'character', 'integer', 'logical', 'complex'] + [*VariablePrecision.real_id] + ['rpe']

        # Stores names of external data type used in declarations
        # Use set() to avoid duplicates
        external_types = set(v.type for v in self.variables if v.type.lower() not in avoid_types)

        # Store the external derived type along with the others
        for external_t in external_types:
            self.derived_types.append(BasicStructures.DerivedType(external_t, None, is_external=True))

    def set_precision_to(self, value):
        """
        Function to loop over all variables where we can change the precision to set the precision value.
        """
        for variable in self.variables:
            if variable.mutable:
                variable.precision = value

    def generate_namelist_precisions(self, filename="namelist_precisions"):
        from os.path import abspath
        # Get the mutable variables (they can only be of rpe type) and sort in ascending ID order
        assigned_rpe_vars = [var for var in self.variables if var.mutable]
        assigned_rpe_vars.sort(key=lambda x: x.id)
        line_text = "emulator_variable_precisions(%d) = %d\t! Variable:%24s\tRoutine:%22s\tModule:%16s\n"
        with open(filename, "w") as namelist:
            namelist.write("! namelist variable precisions\n&precisions\n")
            for var in assigned_rpe_vars:
                # if var.precision == 52 and exclude_double_precision:
                #     continue
                if var.precision is None:
                    continue
                namelist.write(
                    line_text % (var.id, var.precision, var.name, var.procedure.name, var.procedure.module.name))
            namelist.write("/\n")
        print("\nNamelist written to " + abspath(filename))

    def assign_ids_to_immutable_variables(self):
        # Loop over rpe-variables and assign an id to those that don't have any, setting them to mutable=False
        last_id = max([v.id for v in self.variables if v.mutable is True])

        counter = last_id + 1
        for variable in self.variables:
            if variable.type == "rpe_var":
                if variable.id is None:
                    variable.id = counter
                    variable.mutable = False
                    counter += 1

    def save(self, path):
        import sys
        from os.path import abspath
        import pickle
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        sys.setrecursionlimit(10000)
        print("\nVault successfully dumped to ", abspath(path))
        pickle.dump(self, open(path, "wb"))
