# Functions
def hash_from_list(_list):
    """
    Given a list of integers corresponding to variable ids, it generates a unique hash that identifies it
    """
    import sys
    _list = list(sorted(_list))
    val = (sys.maxsize + 1) * 2
    return hash(tuple(_list)) + val


def generate_jobscript(template, parameters):
    """
    Replaces a list of parameters from a provided dictionary in a job template.
    Inputs:
         - template (str): A text job with has some parameters between %% (i.e. %USERNAME%) which will be replaced.
         - parameters (dictionary): A dictionary including keys and values.
    Returns:
         - job (str): text corresponding to a job that can be submited to the compute nodes.
    """
    with open(template) as f:
        job = f.read()
        for parameter, value in parameters.items():
            job = job.replace("%%%s%%" % parameter, value)
    return job


def submit_job(job_text, job_remote_path, communicator):
    """
    Saves a job (text) to a file in the remote machine and submits it to the scheduler.
    """
    import re
    # Creates a new text file in the remote machine and saves the job into it.
    with communicator.sftp.file(job_remote_path, "w") as jobscript:
        jobscript.write(job_text)

    # Defining the scheduler submit command
    # TODO: To use this tool with other schedulers, it might be convenient to define it somewhere else.
    scheduler_submit_command = "sbatch"

    # Defining the command to submit the job
    command = "%s %s" % (scheduler_submit_command, job_remote_path)

    # Executing the command using the communicator
    stdin, stdout, stderr = communicator.execute(command)
    output = str(stdout.read())

    # Read the job id of the submitted job
    job_id_pattern = "Submitted batch job +([0-9]+)"
    job_id = re.search(job_id_pattern, output).group(1)

    # Return job id
    return job_id


def job_status(job_id, communicator):
    """
    Given a job id, it returns the status of the job in the remote machine.
    """
    # Defining the scheduler status command
    # TODO: To use this tool with other schedulers, it might be convenient to define it somewhere else.
    scheduler_status_command = 'sacct  --format="State,Elapsed"'

    # Defining the command that will be sent to the remote machine
    command = '%s --job %s' % (scheduler_status_command, job_id)

    # Executing the command using the communicator
    stdin, stdout, stderr = communicator.execute(command)

    # Read status and time in the output received.
    output = stdout.read().decode()
    try:
        split_output = output.split("\n")[2].split()
        status = split_output[0]
        time = split_output[1]
    except IndexError:
        status = "PENDING"
        time = "00:00"
    # Return status and time
    return status, time


def obtain_result(remote_rundir, communicator, result_name=None):
    """
    Retrieve the results of a simulation from the remote machine.

    As it is right now, we expect the results to be a pickle object, which contents are returned.
    """
    # Import pickle
    import pickle as pkl

    # Define the path in which the result will be located in the remote machine.
    remote_filename = "%s/%s" % (remote_rundir, result_name)

    # Paramiko has severe speed issues if the files that are being opened are bigger than few kBytes,
    # for this reason, we are setting a bigger buffer size in order to avoid this issue.
    buffer_size = 3 * 1024 * 1024
    # Opening a remote file, loading the contents using pickle and returing it
    with communicator.sftp.open(remote_filename, "rb", bufsize=buffer_size) as f:
        return pkl.load(f)


def generate_namelist(variable_set, filename, _vault, reduced_precision_level=10):
    """
    It takes a list of variable ids and generates a RPE namelist which sets the variables's precision to
    the reduced precision level provided as argument.

    """
    # Get the complete list of variables contained in the vault whose precision can be modified
    variables = [v for v in _vault.variables if v.mutable]

    # Loop over all the variables:
    # If he id is in the list of variables, set the precision to the provided one.
    # Otherwise, set it to double precision (52 significant bits )
    for index, v in enumerate(variables):
        if v.id in variable_set:
            v.precision = reduced_precision_level
        else:
            v.precision = 52
    # Create the namelist and save it into the filename provided
    _vault.generate_namelist_precisions(filename)


def var_set_identifier(job):
    """
    Given a BinarySearchJob, it creates a meaningful identifier that can help us to identify an experiment
    """
    to_return = ""
    current_var_set = job

    # Start from the original job, add a character that identifies if the job is the child 0 or 1.
    # Change the current job for the parent and keep doing it until we reach the root of the tree.
    # This will give us a map of where the job is located in the binary tree.
    while current_var_set.parent is not None:
        if len(to_return) > 100:
            raise AssertionError("Something strange happens here!")
        to_return += str(current_var_set.child_number)
        current_var_set = current_var_set.parent
    to_return = "l%02s:%s" % (str(job.level), to_return[::-1])
    return to_return


def divide_set(set0, vault):
    use_old = False
    if use_old:
        return simple_divide_set(set0)
    else:
        return smart_divide(variable_set=set0, vault=vault)


def smart_divide(variable_set, vault):
    variables = [vault.get_variable_by_id(_id) for _id in variable_set]

    # Dividing by modules
    modules = [var.procedure.module.name for var in variables]
    # It is necessary to sort the modules otherwise the order can be different in different runs
    # leading to different groups with different hashes, which prevents reproducibility
    unique_modules = list(sorted(set(modules)))
    if len(unique_modules) > 1:
        d = {}
        for mod in unique_modules:
            d[mod] = modules.count(mod)
        d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
        g1 = {"mod": [], "val": 0}
        g2 = {"mod": [], "val": 0}
        for key, value in d.items():
            if g1["val"] < g2["val"]:
                g1["val"] += value
                g1["mod"].append(key)
            else:
                g2["val"] += value
                g2["mod"].append(key)
        set1 = [v.id for v in variables if v.procedure.module.name in g1["mod"]]
        set2 = [v.id for v in variables if v.procedure.module.name in g2["mod"]]
        return set1, set2
    else:
        # Separate by routines
        routines = [var.procedure.name for var in variables]
        unique_routines = list(sorted(set(routines)))
        if len(unique_routines) > 1:
            d = {}
            for routine in unique_routines:
                d[routine] = routines.count(routine)
            d = {k: v for k, v in sorted(d.items(), key=lambda item: - item[1])}
            g1 = {"mod": [], "val": 0}
            g2 = {"mod": [], "val": 0}
            for key, value in d.items():
                if g1["val"] < g2["val"]:
                    g1["val"] += value
                    g1["mod"].append(key)
                else:
                    g2["val"] += value
                    g2["mod"].append(key)
            set1 = [v.id for v in variables if v.procedure.name in g1["mod"]]
            set2 = [v.id for v in variables if v.procedure.name in g2["mod"]]
            return set1, set2
        else:
            # Just separate in two halves
            return simple_divide_set(variable_set)


def simple_divide_set(set0):
    assert len(set0) > 1
    set1 = []
    set2 = set0[:]
    for i in range(int(len(set0) / 2)):
        set1.append(set2.pop(0))
    return set1, set2
